#!/usr/bin/env python

"""
Module 'controller' serves to connect the view objects with the data objects.
"""

import collections
import datetime
from enum import Enum
import json
import math
import sys

from PyQt5 import QtCore, QtGui, QtWidgets
import networkx as nx
import requests

import data
import enums
import layout
import gui

Coords = collections.namedtuple('Coords', 'x y')

class UpdateWorker(QtCore.QObject):
    """Provides the functionality for network updates."""

    # Signals that may be emitted
    updateFinished = QtCore.pyqtSignal(object)
    connectionClosed = QtCore.pyqtSignal()

    def __init__(self, network, interval, gui):
        """Initialize instance variables.

        Parameters
        ----------
        network : data.NetworkData
            Reference to the object communicating with the network.
        interval : int
            Interval in which updates are performed.
        gui : gui.Window
            Reference to the main windows of the graphical interface.
        """
        super().__init__()
        self.network = network
        self.interval = interval
        self.gui = gui

    def update(self):
        """Request a network update and emit a signal with the changes."""
        try:
            # Request a network update
            changes = self.network.update(self.interval)
        except requests.exceptions.ConnectionError:
            # Evoke a warning message for the caught exception
            self.gui.alert(enums.Alert.connection_closed)
            # Emit a signal that the connection should be closed internally
            self.connectionClosed.emit()
        except requests.exceptions.HTTPError:
            # Evoke a warning message for the caught exception
            self.gui.alert(enums.Alert.connection_lost)
            # Emit a signal that the connection should be closed internally
            self.connectionClosed.emit()
        except:
            # Evoke a warning message for the general exception
            self.gui.alert(enums.Alert.internal_error)
            # Emit a signal that the connection should be closed internally
            self.connectionClosed.emit()
        else:
            # Emit a signal with the changes in the network
            self.updateFinished.emit(changes)
       
class Controller:
    """Defines methods to fill the view and to react to occuring events."""

    def __init__(self, gui):
        """Initialize instance variables and perform initial tasks.

        Parameters
        ----------
        gui : gui.Window
            Reference to the main windows of the graphical interface.
        """
        super().__init__()
        self.gui = gui
        # Set the sidebar to the state of a disconnected network
        self.gui.sidebar.connected(False)
        # Create a new layout working with the given scene
        self.layout = layout.Layout(self.gui.scene)

        # Initialize the object that manages network data and communication
        self.network = data.NetworkData()
        # Set the interval for network updates
        self.update_interval = 10  # in seconds

        # Connect signals from the view with corresponding methods
        self.connect_signals()
        # Prepare a variable that keeps a reference to the network device
        # being investigated in the Detail tab
        self.element_detailed = None

        # Push available network interface modules into the menu in the view
        self.gui_connection_fill_interfaces()
       
        # Initialize internal representation for the path that is being defined
        self.path = Path(self, self.gui.sidebar.path_definition.table)

        # Create a thread for the worker
        self.worker_thread = QtCore.QThread()
        # Create a worker object for network updates
        self.worker = UpdateWorker(self.network, self.update_interval,
            self.gui)
        # Move the worker object to operate in the worker thread
        self.worker.moveToThread(self.worker_thread)
        # Connect signals emitted from the worker object
        self.worker.updateFinished.connect(self.network_update)
        self.worker.connectionClosed.connect(self.network_connection_close)
        # Start the worker thread
        self.worker_thread.start()

    def gui_connection_fill_interfaces(self):
        """Push available network interface modules into the view."""
        for key, name in self.network.options.items():
            self.gui.sidebar.connection.add_interface(name, key)

    def connect_signals(self):
        """Connect signals from the view with corresponding methods."""
        # Sidebar: hide/show button clicked
        self.gui.sidebar.button.clicked.connect(
            self.gui.sidebar.toggle_sidebar)
        # Sidebar: tab changed
        self.gui.sidebar.tabs.currentChanged.connect(
            self.gui_sidebar_tab_changed)

        # Panel Connection: button Connect
        self.gui.sidebar.connection.connect.clicked.connect(
            self.network_connection_open)
        # Panel Connection: button Disconnect
        self.gui.sidebar.connection.disconnect.clicked.connect(
            self.network_connection_close)

        # Panel Path definition: table item double clicked
        self.gui.sidebar.path_definition.table.doubleClicked.connect(
            self.path_remove_from_table)
        # Panel Path definition: button Confirm
        self.gui.sidebar.path_definition.confirm.clicked.connect(
            self.path_confirm)
        # Panel Path definition: button Discard
        self.gui.sidebar.path_definition.discard.clicked.connect(
            self.path_discard)

        # Panel Save/load: button Save
        self.gui.sidebar.saveload.save.clicked.connect(self.network_save)
        # Panel Save/load: button Load
        self.gui.sidebar.saveload.load.clicked.connect(self.network_load)

        # View: datapath level changed
        self.gui.view.datapathLevelChanged.connect(self.network_level_change)
        # View: element moved
        self.gui.view.elementMoved.connect(self.layout.move_element)
        # View: layout changed
        self.gui.view.createTree.connect(self.layout_tree)
        self.gui.view.createGraph.connect(self.layout.graph_renew)
        # View: element clicked
        self.gui.view.elementClicked.connect(self.gui_element_selected)
        # View: element double clicked
        self.gui.view.elementDoubleClicked.connect(self.gui_element_detailed)

        # Defined path: key delete pressed
        self.gui.paths.deletePressed.connect(self.network_remove_path)
    def network_connection_open(self):
        """Open a network connection and perform subsequent tasks."""
        # Validate form fields in the Connection panel.
        if not self.gui.sidebar.connection.check():
            return

        # Extract values from the form fields in the Connection panel.
        interface = (self.gui.sidebar.connection.form_items['interface'].
            currentData())
        ip = self.gui.sidebar.connection.form_items['ip'].text()
        username = self.gui.sidebar.connection.form_items['username'].text()
        password = self.gui.sidebar.connection.form_items['password'].text()

        try:
            # Connect to the network
            self.network.connect(interface, ip, username, password)
        except requests.exceptions.Timeout:
            # Evoke a warning message for the caught exception
            self.gui.alert(enums.Alert.connection_timed_out)
        except:
            # Evoke a warning message for the general exception
            self.gui.alert(enums.Alert.connection_failed)
        else:
            # Set the sidebar to the state of a connected network
            self.gui.sidebar.connected(True)
            # Set the Connection panel to the state of a connected network
            self.gui.sidebar.connection.connected(True)
            # Make Path definition panel the current panel
            self.gui.sidebar.tabs.setCurrentWidget(
                self.gui.sidebar.path_definition)

            # Load network devices from the controller
            net_elements = (self.network.net_controller.datapaths +
                self.network.net_controller.nodes)
            # Load network links from the controller
            net_links = self.network.net_controller.links
            # Create a graph layout from the loaded network data
            self.layout.graph_create(net_elements, net_links)

            # Create a thread for the timer
            self.timer_thread = QtCore.QThread()
            # Create a timer object
            self.timer = QtCore.QTimer()
            # Connect the timeout signal with the update method of UpdateWorker
            self.timer.timeout.connect(self.worker.update)
            # Start the timer to timeout in the given interval
            # Interval has to be given in miliseconds
            self.timer.start(self.update_interval * 1000)
            # Move the timer object to operate in the timer thread
            self.timer.moveToThread(self.timer_thread)
            # Start the timer thread
            self.timer_thread.start()

    def network_connection_close(self):
        """Close the network connection."""
        # Quit the timer thread
        self.timer_thread.quit()

        # Disconnect from the network
        self.network.disconnect()
        # Set the sidebar to the state of a disconnected network
        self.gui.sidebar.connected(False)
        # Set the Connection panel to the state of a disconnected network
        self.gui.sidebar.connection.connected(False)

        # Clear the layout
        self.layout.clear()

    def network_update(self, changes):
        """Process changes in the network.

        Parameters
        ----------
        changes : interfaces.Update
            A summary of changes in the network.
        """
        # Iterate network devices to be removed
        for net_element in changes.remove:
            # In case the element was detailed in the Detail panel
            # Set the Detail panel empty
            if self.element_detailed is net_element:
                self.gui.sidebar.detail.set_empty()

            # If the device was a datapath that was selected in the path
            # definition
            if (self.path.datapaths and 
                    isinstance(net_element, data.interfaces.Datapath) and
                    net_element in self.path.datapaths):
                # Evoke a warning message
                self.gui.alert(enums.Alert.datapath_inactive)
                # Clear the path
                self.path_clear()
            # If the device was a node that was selected in the path definition
            elif (self.path.nodes and
                    isinstance(net_element, data.interfaces.Node) and
                    net_element in self.path.nodes):
                # Evoke a warning message
                self.gui.alert(enums.Alert.node_inactive)
                # Clear the path
                self.path_clear()
                
        # Update the layout
        self.layout.update(changes.add, changes.remove,
                           changes.connect, changes.disconnect)
        
        # Update the state about hierarchy levels in case of removed datapaths
        # Influences whether a tree layout may be created
        self.gui.view.network_levels_state_changed(
            self.network.datapath_levels_state())

    def network_level_change(self, grph_element, old_level, new_level):
        """Change the hierarchy level for the given datapath.

        Parameters
        ----------
        grph_element : gui.canvas.Datapath
            Graphical element of the datapath to change.
        old_level : enums.DeviceType
            Old level of the datapath.
        new_level : enums.DeviceType
            New level of the datapath.
        """
        # Obtain a network element for the given graphical element
        net_element = self.layout.translation.element_from_graphical[
           grph_element]
        # Change the level in the network
        self.network.datapath_level_change(net_element, old_level, new_level)

        # Change the level in the graphical element
        grph_element.set_level(new_level)
        # Update the state about hierarchy levels
        # Influences whether a tree layout may be created
        self.gui.view.network_levels_state_changed(
            self.network.datapath_levels_state())
        # Change the level in the table overview for path definition
        self.path.change_datapath_level(net_element, new_level)

    def network_name_change(self, name):
        """Change the name for the given device.

        Parameters
        ----------
        name : string
            Name to set.
        """
        # Get a unique identifier of the detailed device
        # DPID for a datapath
        if isinstance(self.element_detailed, data.interfaces.Datapath):
            identifier = self.element_detailed.dpid
        # MAC address for a node
        elif isinstance(self.element_detailed, data.interfaces.Node):
            identifier = self.element_detailed.mac
        
        # If the name is empty, the identifier is set as the name
        if name == "":
            name = identifier
        # Change the name in the network
        data.NetworkData.device_names[identifier] = name
        
        # Obtain a graphical element for the network element
        grph_element = self.layout.translation.element_to_graphical[
            self.element_detailed]
        # Change the name in the graphical element
        grph_element.rename(name)

    def layout_tree(self):
        """Change the layout to a tree-like structure."""
        # Obtain live datapaths
        res = self.network.datapath_levels_alive()
        # The tree can be created only if there are some
        if res is not None:
            level, net_elements = res
            self.layout.tree(net_elements)

    def gui_element_detailed(self, grph_element):
        """Set the chosen device to be detailed in the Detail panel.

        Parameters
        ----------
        grph_element : gui.canvas.Datapath or gui.canvas.Node
            Graphical element of the device to be detailed.
        """
        # Obtain a network element for the given graphical element
        net_element = self.layout.translation.element_from_graphical[
            grph_element]
        # Set the network element for the detailed device
        self.element_detailed = net_element

        # In case it is a datapath
        if isinstance(net_element, data.interfaces.Datapath):
            # Prepare information to present
            dpid = net_element.dpid
            ip = net_element.ip
            of_version = net_element.of_version

            # Obtain the name of the device
            name = (data.NetworkData.device_names[dpid] if dpid in
                data.NetworkData.device_names else dpid)
            # Push the information into the Detail panel
            self.gui.sidebar.detail.set_datapath(name, dpid, ip, of_version)
        # In case it is a node
        elif isinstance(net_element, data.interfaces.Node):
            # Prepare information to present
            mac = net_element.mac
            ip = net_element.ip
            parent_dpid = net_element.parent.dpid
            parent_ip = net_element.parent.ip
            parent_port = net_element.parent_port

            # Obtain the name of the device
            name = (data.NetworkData.device_names[mac] if mac in
                data.NetworkData.device_names else mac)
            # Push the information into the Detail panel
            self.gui.sidebar.detail.set_node(name, mac, ip,
                                     parent_dpid, parent_ip, parent_port)
        # Set the Detail panel as the current panel
        self.gui.sidebar.tabs.setCurrentWidget(self.gui.sidebar.detail)
        # Show the sidebar in case it was hidden
        self.gui.sidebar.toggle_sidebar(True)
        # Connect the signal of a changed name
        # The field for the name is newly created for every detailed device,
        # therefore the signal always has to be connected again
        self.gui.sidebar.detail.name.textEdited.connect(
            self.network_name_change)

    def gui_element_selected(self, grph_element):
        """Add the chosen device into or remove from the defined path.

        Parameters
        ----------
        grph_element : gui.canvas.Datapath or gui.canvas.Node
            Graphical element of the device to be detailed.
        """
        # Obtain a network element for the given graphical element
        net_element = self.layout.translation.element_from_graphical[
            grph_element]
        # If the device is already selected, remove it from the path
        if grph_element.selected:
            self.path_remove_element(net_element)
        # In case it is not selected and the path is not yet complete,
        # add it to the path
        elif not grph_element.selected and not self.path.completed:
            result = self.path.add_element(net_element)
            # Graphically mark the element as selected if it was
            # sucessfully added to the path
            if result is True:
                grph_element.select()
            # In case of trouble, evoke a warning message
            else:
                self.gui.alert(result)

    def gui_sidebar_tab_changed(self, tab_index):
        """Turn path definition on or off depeneding on the chosen panel.
 
        Parameters
        ----------
        tab_index : int
            Index number of the current panel.
        """
        # If the current panel is Path Definition, turn the functionality on
        if isinstance(self.gui.sidebar.tabs.widget(tab_index),
                      gui.panels.PathDefinition):
            self.gui.view.path_definition_state_changed(True)
        # If the current panel is not Path Definition,
        # turn the functionality off
        else:
            self.gui.view.path_definition_state_changed(False)

    def path_remove_element(self, net_element):
        """Remove the given network device from the defined path.

        Parameters
        ----------
        net_element : interfaces.Datapath or interfaces.Node
            Network device to remove from the defined path.
        """
        # Remove the given network device from the path
        result = self.path.remove_element(net_element)
        # If result is a list of elements to graphically deselect
        if isinstance(result, list):
            self.path_deselect_elements(result)
        # In case the removal has to be confirmed through a dialog message
        else:
            answer = self.gui.alert_question(result)
            if answer:
                result_confirmed = self.path.remove_element(
                    net_element, confirmed=result)
                self.path_deselect_elements(result_confirmed)

    def path_deselect_elements(self, net_elements):
        """Graphically deselect given elements.

        Parameters
        ----------
        net_elements : list of intertfaces.Datapath or interfaces.Node
            A list of network devices to be deselected in the visualization.
        """
        for net_element in net_elements:
            # Obtain a graphical element for the given network element
            grph_element = (self.layout.translation.element_to_graphical[
                net_element])
            # Deselect the graphical element
            grph_element.deselect()

    def path_remove_from_table(self, index):
        """Remove a part of the path based on element selected in the table.

        Parameters
        ----------
        index : int
            Index of the double clicked item.
        """
        # Obtain the network element on the given row
        net_element = self.path.element_on_row(index.row())
        # Remove the element
        self.path_remove_element(net_element)

    def path_complete_state_changed(self, state):
        """Take actions when the path is complete or made incomplete.

        Parameters
        ----------
        state : bool
            True if the path is complete, False otherwise.
        """
        self.gui.sidebar.path_definition.path_complete_state_changed(state)

    def path_confirm(self):
        """Confirm the path definition."""
        # Validate form fields in the Path definition panel
        if not  self.gui.sidebar.path_definition.check():
            return

        # Gather data of the path
        # Source end node
        src = self.path.nodes[0]
        # Destination end node
        dst = self.path.nodes[1]
        # List of involved datapaths
        datapaths = self.path.datapaths
        # Values defined in the form in the Path definition panel
        values = self.gui.sidebar.path_definition.values()

        try:
            # Push the path to the network
            path = self.network.add_path(src, dst, datapaths, values[0], values[1])
        except:
            # Evoke a warning message for the general exception
            self.gui.alert(enums.Alert.path_definition_failed)
        else:
            # Clear the path definition
            self.path_clear()
            # Add the path to the Defined paths widget
            self.gui_defined_paths_add(path)

    def path_discard(self):
        """Discard the defined path."""
        # A confirmation is required through a question dialog
        if (self.path.nodes and
                self.gui.alert_question(enums.AlertQuestion.discard_defined_path)):
            self.path_clear()

    def path_clear(self):
        """Clear the path definition."""
        # Graphically deselect all elements of the path
        self.path_deselect_elements(self.path.clear())
        # Clear the form in the Path definition panel
        self.gui.sidebar.path_definition.clear()

    def gui_defined_paths_add(self, path):
        """Add a defined path to the Defined paths widget.

        Parameters
        ----------
        path : data.DefinedPath
            Defined path to add.
        """
        # Create a tree structure of the path parameters
        tree = self.gui.paths.tree
        name = QtWidgets.QTreeWidgetItem([path.name])
        match = QtWidgets.QTreeWidgetItem(name, ["Match fields"])
        src_ip = QtWidgets.QTreeWidgetItem(match,
            ["Source IP address: "+path.match['src_ip']])
        dst_ip = QtWidgets.QTreeWidgetItem(match,
            ["Destination IP address: "+path.match['dst_ip']])

        if path.match['ip_protocol'] is enums.IpProtocol.tcp:
            ip_protocol = QtWidgets.QTreeWidgetItem(match,
                ["IP protocol: TCP"])
            if 'tcp_src' in path.match:
                tcp_src = QtWidgets.QTreeWidgetItem(match,
                    ["TCP source port: "+str(path.match['tcp_src'])])
            if 'tcp_dst' in path.match:
                tcp_dst = QtWidgets.QTreeWidgetItem(match,
                    ["TCP destination port: "+str(path.match['tcp_dst'])])
        elif path.match['ip_protocol'] is enums.IpProtocol.udp:
            ip_protocol = QtWidgets.QTreeWidgetItem(match,
                ["IP protocol: UDP"])
            if 'udp_src' in path.match:
                udp_src = QtWidgets.QTreeWidgetItem(match,
                    ["UDP source port: "+str(path.match['udp_src'])])
            if 'udp_dst' in path.match:
                udp_dst = QtWidgets.QTreeWidgetItem(match,
                    ["UDP destination port: "+str(path.match['udp_dst'])])

        datapaths = QtWidgets.QTreeWidgetItem(name, ["Datapaths"])
        for dp in path.datapaths:
            dpid = QtWidgets.QTreeWidgetItem(datapaths, [dp['dpid']])
            of_version = QtWidgets.QTreeWidgetItem(dpid,
                ["OpenFlow version: "+dp['of_version']])
            in_port = QtWidgets.QTreeWidgetItem(dpid,
                ["Input port: "+str(dp['in_port'])])
            out_port = QtWidgets.QTreeWidgetItem(dpid,
                ["Output port: "+str(dp['out_port'])])

        priority = QtWidgets.QTreeWidgetItem(name,
            ["Priority: "+str(path.priority)])
        idle_timeout = QtWidgets.QTreeWidgetItem(name,
            ["Idle timeout: "+str(path.idle_timeout)])
        hard_timeout = QtWidgets.QTreeWidgetItem(name,
            ["Hard timeout: "+str(path.hard_timeout)])
        time = datetime.datetime.fromtimestamp(path.time)
        date = QtWidgets.QTreeWidgetItem(name,
            ["Date of definition: "+time.strftime('%Y/%m/%d %H:%M:%S')])

        # Add the tree to the widget
        tree.addTopLevelItem(name)

    def network_remove_path(self, row):
        """Remove a defined path from the local storage.

        Parameters
        ----------
        row : int
            Row number of the path to remove.
        """
        # Remove the path from the Defined paths widget
        self.gui.paths.tree.takeTopLevelItem(row)
        # Remove the path from the NetworkData object
        del data.NetworkData.defined_paths[row]

    def network_save(self):
        """Save network data to an external file."""
        # Obtain the filename through the graphical interface
        filename = self.gui.save_file()[0]
        if not filename:
            return

        # Open the file for writing
        with open(filename, 'w') as f:
            # Obtain data to export
            data = self.export_data()
            # Write a JSON representation of the data into the file
            json.dump(data, f, cls=EnumEncoder)

    def network_load(self):
        """Load network data from an external file."""
        # Obtain the filename through the graphical interface
        filename = self.gui.open_file()[0]
        if not filename:
            return

        # Open the file for reading
        with open(filename, 'r') as f:
            # Load data from the JSON representation from the file
            data = json.load(f, object_hook=as_enum)
            # Import the loaded data
            self.import_data(data)

    def export_data(self):
        """Gather network data for export.

        Returns
        -------
        dict
            Dictionary with network data (hierarchy levels, names, layout
            positions) and defined paths.
        """
        # Initialize a dictionary for network data
        network = {'datapaths': {}, 'nodes': {}}

        # Iterate elements to export
        for grph_element in self.layout.positions:
            # Obtain a network element for the given graphical element
            net_element = (self.layout.translation.
                element_from_graphical[grph_element])
            # If the element is a datapath
            if isinstance(net_element, data.interfaces.Datapath):
                network['datapaths'][net_element.dpid] = {}
                # Add hierarchy level
                network['datapaths'][net_element.dpid]['level'] = (data.
                    Network.datapath_level(net_element))
                # Add layout position
                network['datapaths'][net_element.dpid]['position'] = tuple(
                    self.layout.positions[grph_element])
                # Add name if defined for this datapath
                if net_element.dpid in data.NetworkData.device_names:
                    network['datapaths'][net_element.dpid]['name'] = (data.
                        Network.device_names[net_element.dpid])
            # If the element is a node
            elif isinstance(net_element, data.interfaces.Node):
                network['nodes'][net_element.mac] = {}
                # Add layout position
                network['nodes'][net_element.mac]['position'] = tuple(
                    self.layout.positions[grph_element])
                # Add name if defined for this node
                if net_element.mac in data.NetworkData.device_names:
                    network['nodes'][net_element.mac]['name'] = (data.
                        Network.device_names[net_element.mac])

        # Initialize a list for defined paths
        paths = []
        for path in data.NetworkData.defined_paths:
            # Add to the list a dictionary of variables of the path object
            paths.append(path.__dict__)

        # Compile the return dictionary
        mashup = {'network': network, 'paths': paths}
        return mashup

    def import_data(self, mashup):
        """Import network data.

        Parameters
        ----------
        mashup : dict
            Dictionary with network data and defined paths.
        """
        # Iterate nodes to import
        for mac, node in mashup['network']['nodes'].items():
            # Obtain a Node object for the given MAC address
            net_element = self.network.net_controller.node_by_mac(mac)
            # If the node is alive (Node object exists)
            if net_element is not None:
                # Obtain a graphical element for the given network element
                grph_element = self.layout.translation.element_to_graphical[
                    net_element]
                # Position the element in the layout
                self.layout.positions[grph_element] = node['position']
                grph_element.setPos(node['position'][0], node['position'][1])
            # If the node has defined name, rename it
            if 'name' in node:
                data.NetworkData.device_names[mac] = node['name']
                if grph_element:
                    grph_element.rename(node['name'])

        # Iterate datapaths to import
        for dpid, datapath in mashup['network']['datapaths'].items():
            # Obtain a Datapath object for the given DPID
            net_element = self.network.net_controller.datapath_by_dpid(dpid)
            # If the datapath is alive (Datapath object exists)
            if net_element is not None:
                # Obtain a graphical element for the given network element
                grph_element = self.layout.translation.element_to_graphical[
                    net_element]
                # Position the element in the layout
                self.layout.positions[grph_element] = datapath['position']
                grph_element.setPos(datapath['position'][0],
                                    datapath['position'][1])
            # If the datapath has defined name, rename it
            if 'name' in datapath:
                data.NetworkData.device_names[mac] = datapath['name']
                if grph_element:
                    grph_element.rename(datapath['name'])
            # If the hierarchy level is core or distribution,
            # add the datapath to the corresponding set
            if datapath['level'] in data.NetworkData.levels:
                data.NetworkData.levels[datapath['level']].add(dpid)
                # If the datapath is alive, change the level in its
                # graphical element
                if grph_element:
                    grph_element.set_level(datapath['level'])
                    # Update the state about hierarchy levels
                    # Influences whether a tree layout may be created
                    self.gui.view.network_levels_state_changed(
                        self.network.datapath_levels_state())

        # Redraw graphical links because of the position changes
        for grph_link in self.layout.translation.link_to_graphical.values():
            grph_link.draw()

        # Iterate paths to import
        for path in mashup['paths']:
            # Create an object for the local representation of the defined path
            new = data.DefinedPath(path['name'], path['match']['src_ip'],
                path['match']['dst_ip'], path['priority'],
                path['idle_timeout'], path['hard_timeout'], path['time'])
        
            # Process optional fields
            opt_match = {}
            opt_match['ip_protocol'] = path['match']['ip_protocol']
            if path['match']['ip_protocol'] is enums.IpProtocol.tcp:
                if 'tcp_src' in path['match']:
                    opt_match['tcp_src'] = path['match']['tcp_src']
                if 'tcp_dst' in path['match']:
                    opt_match['tcp_dst'] = path['match']['tcp_dst']
            elif path['match']['ip_protocol'] is enums.IpProtocol.udp:
                opt_match['ip_protocol'] = 'udp'
                if 'udp_src' in path['match']:
                    opt_match['udp_src'] = path['match']['udp_src']
                if 'udp_dst' in path['match']:
                    opt_match['udp_dst'] = path['match']['udp_dst']
            # Add the optional fields to the path
            new.match.update(opt_match)

            # Iterate datapaths involved in the path
            for datapath in path['datapaths']:
                # Add the datapath to the path
                new.add_datapath(datapath['dpid'], datapath['of_version'],
                                 datapath['in_port'], datapath['out_port'])

            # Add the path to NetworkData
            data.NetworkData.defined_paths.append(new)
            # Add the path to the Defined paths widget
            self.gui_defined_paths_add(new)

class EnumEncoder(json.JSONEncoder):
    """Custom encoder to transform Enum objects into the JSON format."""

    def default(self, obj):
        """Returns a serializable object for the given object.

        Parameters
        ----------
        obj : object
            Object to be serialized.

        Returns
        -------
        string
            Serialized representation of the given object.
        """
        # If the object is Enum object
        if isinstance(obj, Enum):
            # Return the name of the object type 
            return {"__enum__": str(obj)}
        # For other objects use base implementation
        return json.JSONEncoder.default(self, obj)

def as_enum(d):
    """Return Enum object for its string representation.

    Parameters
    ----------
    d : dict
        Dictionary object.

    Returns
    -------
    Enum
        In case the dictionary value is a representation of Enum.
    dict
        In case the dictionary value is not a representation of Enum.
    """
    # Compare the dictionary key
    if "__enum__" in d:
        # name = Enum class
        # member = Enum value
        name, member = d["__enum__"].split(".")
        # Obtain a reference for the module enums
        get_module = __import__('enums', fromlist=[name])
        # Obtain a reference for the class of the given module
        # for the given name
        get_class = getattr(get_module, name)
        # Obtain a reference for the member value of the given class
        # for the given name
        attribute = getattr(get_class, member)
        return attribute
    # If not Enum return the original
    else:
        return d


#path manipuluje s prave rozpracovanou cestu
#path table je datove pozadi tabulky v bocnim panelu, ktera obsahuje pridane
#prvky
class Path:
    """Keeps track of the path that is being defined."""
    def __init__(self, controller, table):
        """Initialize instance variables.

        Parameters
        ----------
        controller : Controller
            Reference to Controller.
        table : PyQt5.QtWidgets.QTableWidget
            Reference to the table in the Path definition panel.
        """
        super().__init__()
        # Initialize lists of devices in the path
        self.nodes = []
        self.datapaths = []

        self.controller = controller
        self.table = table

        # Initial state of the path definition
        self.completed = False

    def change_datapath_level(self, net_element, new_level):
        """Change the hierarchy level of a datapath in the path.

        Parameters
        ----------
        net_element : interfaces.Datapath
            Datapath to change.
        new_level : enums.DeviceType
            New hierarchy level to set.
        """
        # If the datapath is in the path
        if net_element in self.datapaths:
            # Get row index of the datapath
            row = self.datapaths.index(net_element) + 1
            # Remove the field with the icon from the table
            item = self.table.takeItem(row, 0)
            # Set the field with a new icon
            icon = new_level
            item.setIcon(QtGui.QIcon(QtGui.QPixmap(
                    gui.canvas.RES_IMAGES+gui.canvas.ICONS[new_level])))
            # Put the altered field back in the table
            self.table.setItem(row, 0, item)

    def add_to_table(self, net_element):
        """Add a network device to the path definition table.

        Parameters
        ----------
        net_element : interfaces.Datapath or interfaces.Node
            Network device to add.
        """
        # Insert a new row in the table
        r = self.table.rowCount()
        self.table.insertRow(r)
 
        # Prepare corresponding icon and unique identifier (DPID or MAC
        # address)
        if isinstance(net_element, data.interfaces.Datapath):
            icon = data.NetworkData.datapath_level(net_element)
            uid = net_element.dpid
        elif isinstance(net_element, data.interfaces.Node):
            icon = enums.DeviceType.node
            uid = net_element.mac

        # Add the icon to the unique identifier field
        uid = QtWidgets.QTableWidgetItem(uid)
        uid.setIcon(QtGui.QIcon(QtGui.QPixmap(
                    gui.canvas.RES_IMAGES+gui.canvas.ICONS[icon])))

        # Fill the row fields (unique identifier and IP address)
        self.table.setItem(r, 0, QtWidgets.QTableWidgetItem(uid))
        self.table.setItem(r, 1, QtWidgets.QTableWidgetItem(net_element.ip))

    def element_on_row(self, row):
        """Get the element placed on the given row.

        Parameters
        ----------
        row : int
            Row number.

        Returns
        -------
        interfaces.Datapath
            In case a datapath is placed on the row.
        interfaces.Node
            In case a node is placed on the row.
        None
            In case the index is out of range.
        """
        # Count index of the last row
        last = len(self.datapaths) + 1
        # Source end node
        if row == 0:
            return self.nodes[0]
        # Destination end node
        if row == last:
            return self.nodes[1]
        # Datapath between the end nodes
        if row > 0 and row < last:
            return  self.datapaths[row-1]
        else:
            return None

    def add_element(self, net_element):
        """Add a network device to the path.

        Parameters
        ----------
        net_element : interfaces.Datapath or interfaces.Node
            Network device to add.

        Returns
        -------
        bool
            True if the device has been successfully added.
        enums.Alert
            In case the device could not have been added for a certain reason.
        """
        # If the device is a node
        if isinstance(net_element, data.interfaces.Node):
            # Add as a source end node in case it is not defined
            if not self.nodes:
                self.nodes.append(net_element)
                self.add_to_table(net_element)
                return True
            # A datapath has to be between two nodes
            elif not self.datapaths:
                return enums.Alert.datapath_before_end_node
            # Add as a destination end node
            else:
                last_datapath = self.datapaths[-1]
                # It is not connected to the last datapath in the path
                if not last_datapath.has_connected(net_element):
                    return enums.Alert.selected_must_neighbor
                # It can be added
                else:
                    self.nodes.append(net_element)
                    self.add_to_table(net_element)
                    self.complete(True)
                    return True
        # If the device is a datapath
        elif isinstance(net_element, data.interfaces.Datapath):
            # There has to be the source end node
            if not self.nodes:
                return enums.Alert.first_node_first
            else:
                # The datapath has to be connected to the last datapath
                previous = (self.datapaths[-1] if self.datapaths
                                               else self.nodes[0])
                if net_element.has_connected(previous):
                    self.datapaths.append(net_element)
                    self.add_to_table(net_element)
                    return True
                else:
                    return enums.Alert.selected_must_neighbor

    def remove_element(self, net_element, confirmed=False):
        """Remove a network device from the path.

        Parameters
        ----------
        net_element : interfaces.Datapath or interfaces.Node
            Network device to remove.
        confirmed : bool or int
            Whether a question dialog has been confirmed or not.
            If yes, it contains the Enum numeric value.

        Returns
        -------
        list
            List of devices to graphically remove from the path.
        enums.AlertQuestion
            In case the removal has to be confirmed. 
        """
        # If the network device is a node
        if isinstance(net_element, data.interfaces.Node):
            # The node is the source end node
            if self.nodes[0] is net_element:
                # Directly remove in case there are no datapaths
                if not self.datapaths:
                    to_remove = [self.nodes[0]]

                    self.table.removeRow(0)
                    del(self.nodes[0])

                    return to_remove
                # Ask for confirmation to remove the node and all succeeding
                # devices
                else:
                    if confirmed == enums.AlertQuestion.start_all_over:
                        return self.clear()
                    else:
                        return enums.AlertQuestion.start_all_over
            # The node is the destination end node
            # Remove directly
            elif self.nodes[1] is net_element:
                to_remove = [self.nodes[1]]

                r = self.table.rowCount()
                self.table.removeRow(r-1)
                del(self.nodes[1])

                self.complete(False)
                return to_remove
        # If the network device is a datapath
        elif isinstance(net_element, data.interfaces.Datapath):
            # In case the datapath is the last datapath and the destination
            # end node is not defined
            # Remove directly
            if self.datapaths[-1] is net_element and len(self.nodes) == 1:
                to_remove = [self.datapaths[-1]]

                r = self.table.rowCount()
                self.table.removeRow(r-1)
                del(self.datapaths[-1])

                return to_remove
            # Ask for confirmation to remove the datapath and all
            # succeeding devices
            else:
                if confirmed == enums.AlertQuestion.delete_all_succeeding:
                    index = self.datapaths.index(net_element)
                    to_remove = self.nodes[1:] + self.datapaths[index:]

                    while self.table.rowCount() > (1 + index):
                        self.table.removeRow(1 + index)

                    del(self.nodes[1:])
                    del(self.datapaths[index:])

                    self.complete(False)
                    return to_remove
                else:
                    return enums.AlertQuestion.delete_all_succeeding

    def clear(self):
        """Clear the defined path.

        Returns
        -------
        list
            List of devices to graphically remove from the path.
        """
        to_remove = self.nodes + self.datapaths

        self.nodes = []
        self.datapaths = []
        while self.table.rowCount() > 0:
            self.table.removeRow(0)

        self.complete(False)
        return to_remove

    def complete(self, state):
        """Change the state of the path definition.

        Parameters
        ----------
        state : bool
            Whether the path is complete (True) or incomplete (False).
        """
        self.completed = state
        self.controller.path_complete_state_changed(state)
