#!/usr/bin/env python

"""
Module 'data' is concerned with the connected network and related data.
"""

from collections import OrderedDict

import datetime
import enums
import interfaces

class NetworkData:
    """Provides network connection and stores custom network information."""

    # Dictionary of pairs enums.DeviceType: set of DPIDs
    levels = {enums.DeviceType.datapath_core: set(),
              enums.DeviceType.datapath_distribution: set()}
    # Dictionary of pairs DPID/MAC: name
    device_names = {}
    # List of DefinedPath objects
    defined_paths = []

    @classmethod
    def datapath_level(cls, datapath):
        """Return the hierarchy level for the given datapath.

        Parameters
        ----------
        datapath : interfaces.Datapath
            Datapath to inquire.

        Returns
        -------
        enums.DeviceType
            Corresponding device type for the given datapath.
        None
            In case the given object is not a datapath.
        """
        if not isinstance(datapath, interfaces.Datapath):
            return None

        # If the datapath is distribution
        if (datapath.dpid in
                cls.levels[enums.DeviceType.datapath_distribution]):
            return enums.DeviceType.datapath_distribution
        # If the datapath is core
        elif datapath.dpid in cls.levels[enums.DeviceType.datapath_core]:
            return enums.DeviceType.datapath_core
        # If the datapath is access
        else:
            return enums.DeviceType.datapath_access

    @classmethod
    def datapath_level_change(cls, datapath, old_level, new_level):
        """Change the hierarchy level of a datapath.

        Parameters
        ----------
        datapath : interfaces.Datapath
            Datapath to change
        old_level : enums.DeviceType
            Old level of the datapath.
        new_level : enums.DeviceType
            New level of the datapath.
        """

        if not isinstance(datapath, interfaces.Datapath):
            return

        # If the old level is core or distribution, remove the datapath
        # from the corresponding set
        if old_level in cls.levels:
            cls.levels[old_level].remove(datapath.dpid)
        # If the new level is core or distribution, add the datapath
        # to the corresponding set
        if new_level in cls.levels:
            cls.levels[new_level].add(datapath.dpid)

    def __init__(self):
        """Initialize instance variables."""
        self.net_controller = None

    @property
    def options(self):
        """Get a dictionary of available controller interfaces.

        Returns
        -------
        OrderedDict
            A dictionary with pairs - module name: full name.
            Ordered by the full name.
        """
        return OrderedDict(sorted(
            interfaces.listing.items(), key=lambda l: l[1]))

    def connect(self, interface, ip, username, password):
        """Connect to a network controller using the given interface.

        Parameters
        ----------
        interface : string
            Module name of the controller interface to connect through.
        ip : string
            IP address of the controller server.
        username : string
            Username to log in to the controller.
        password : string
            Password to log in to the controller.
        """
        self.net_controller = interfaces.create(interface)
        self.net_controller.connect(ip, username, password)
        self.net_controller.load()

    def disconnect(self):
        """Disconnect from the controller by losing a reference to it."""
        self.net_controller = None

    def update(self, interval):
        """Evoke network update and return the result.

        Parameters
        ----------
        interval : int
            The interval between individual updates.

        Returns
        -------
        interfaces.Update
            A summary of changes in the network.
        """
        return self.net_controller.update(interval)

    def datapath_levels_state(self):
        """Check whether any core or distribution datapaths are alive.

        Returns
        -------
        bool
            True in case there is at least one core or distribution datapath
            alive. False otherwise.
        """
        # Iterate DPIDs of core datapaths
        for core in self.levels[enums.DeviceType.datapath_core]:
           # If the network interface returns a Datapath object for the given
           # DPID, the datapath is alive
           if self.net_controller.datapath_by_dpid(core) is not None:
                return True

        # Iterate DPIDs of distribution datapaths
        for dist in self.levels[enums.DeviceType.datapath_distribution]:
           # If the network interface returns a Datapath object for the given
           # DPID, the datapath is alive
           if self.net_controller.datapath_by_dpid(dist) is not None:
                return True

        return False

    def datapath_levels_alive(self):
        """Get live datapaths from the top of the hierarchy.

        Returns
        -------
        list of interfaces.Datapath
            A list of live core datapaths in case any of them are alive.
            A list of live distribution datapaths in case any of them are
            alive and no core datapaths are alive.
        None
            In case no core or distribution datapaths are alive.
        """
        # Prepare an empty list of live datapaths
        alive = []

        # Iterate DPIDs of core datapaths
        for core in self.levels[enums.DeviceType.datapath_core]:
            # Obtain an object for this datapath
            datapath = self.net_controller.datapath_by_dpid(core)
            # In case it is alive, add it to the list
            if datapath is not None:
                alive.append(datapath)

        # If any core datapath are alive, return the list
        if alive:
            return (enums.DeviceType.datapath_core, alive)

        # Iterate DPIDs of distribution datapaths
        for dist in self.levels[enums.DeviceType.datapath_distribution]:
            # Obtain an object for this datapath
            datapath = self.net_controller.datapath_by_dpid(dist)
            # In case it is alive, add it to the list
            if datapath is not None:
                alive.append(datapath)

        # If any distribution datapath are alive, return the list
        # Return None otherwise
        return ((enums.DeviceType.datapath_distribution, alive) if alive
            else None)

    def add_path(self, src, dst, datapaths, mandatory, optional):
        """Program the given path into the connected network

        Parameters
        ----------
        src : string
            IP address of the source end node.
        dst : string
            IP address of the destination end node.
        datapaths : list of interfaces.Datapath
            A list of datapaths into which flow entries are to be installed.
        mandatory : dictionary
            A dictionary of elementary parameters of the path.
            Keys: name, priority, idle_timeout, hard_timeout.
        optional : dictionary
            A dictionary of optional parameters of the path.
            Possible keys: ip_protocol, tcp_src, tcp_dst, udp_src, udp_dst.

        Returns
        -------
        DefinedPath
            Internal representation of the defined path.
        """
        # Initialize the object for the internal representation of the path
        path = DefinedPath(mandatory['name'], src.ip, dst.ip,
            mandatory['priority'], mandatory['idle_timeout'],
            mandatory['hard_timeout'])

        # Check the optional parameters and add them to the internal path
        # in case they are defined
        opt_match = {}
        opt_match['ip_protocol'] = optional['ip_protocol']
        if optional['ip_protocol'] is enums.IpProtocol.tcp:
            if optional['tcp_src']:
                opt_match['tcp_src'] = optional['tcp_src']
            if optional['tcp_dst']:
                opt_match['tcp_dst'] = optional['tcp_dst']
        elif optional['ip_protocol'] is enums.IpProtocol.udp:
            opt_match['ip_protocol'] = 'udp'
            if optional['udp_src']:
                opt_match['udp_src'] = optional['udp_src']
            if optional['udp_dst']:
                opt_match['udp_dst'] = optional['udp_dst']
        path.match.update(opt_match)

        dp_count = len(datapaths)
        # Set the source end node as the previous device for the first datapath
        pred = src
        # Iterate datapaths involved in the path
        for index in range(dp_count):
            curr = datapaths[index]
            # Check if this datapaths is the last one
            # If yes than the next device is the destination end node
            if index == dp_count-1:
                succ = dst
            else:
                succ = datapaths[index+1]
            
            # Input port = port where the previous device is connected
            in_port = curr.port_of_device(pred)
            # Output port = port where the next device is connected
            out_port = curr.port_of_device(succ)

            # Send the flow entry definition to the network interface
            self.net_controller.add_flow(curr, src.ip, dst.ip,
                in_port, out_port, mandatory['priority'],
                mandatory['idle_timeout'], mandatory['hard_timeout'],
                **optional)

            # Add the datapath to the internal representation
            path.add_datapath(curr.dpid, curr.of_version,
                in_port, out_port)

            # Set the current datapath as previous
            pred = curr

        # Add the complete path to the local list of defined paths
        self.defined_paths.append(path)

        # Return the internal representation of the defined path
        return path

class DefinedPath:
    """A data container for a custom user-defined traffic path."""

    def __init__(self, name, src_ip, dst_ip, priority, idle_timeout,
                 hard_timeout, time=None):
        """Initialize instance variables.

        Parameters
        ----------
        name : string
            User-defined name of the path.
        src_ip : string
            IP address of the source end node.
        dst_ip : string
            IP address of the destination end node.
        priority : int
            Priority of the flow entries.
        idle_timeout : int
            Idle timeout for the flow entries.
        hard_timeout : int
            Hard timeout for the flow entries.
        time : int
            Timestamp of when the path was defined.
        """
        self.name = name
        self.match = {'src_ip': src_ip, 'dst_ip': dst_ip}
        self.datapaths = []
        self.priority = priority
        self.idle_timeout = idle_timeout
        self.hard_timeout = hard_timeout
        if time is None:
            now = datetime.datetime.now()
            self.time = now.timestamp()
        else:
            self.time = time

    def add_datapath(self, dpid, of_version, in_port, out_port):
        """Add a datapath to the defined path.

        Parameters
        ----------
        dpid : string
            DPID of the datapath.
        of_version : string
            OpenFlow version of the datapath.
        in_port : int
            Input port at which it matches packets.
        out_port : int
            Output port through which it sends out forwarded packets.
        """
        datapath = {'dpid': dpid, 'of_version': of_version,
                    'in_port': in_port, 'out_port': out_port}
        self.datapaths.append(datapath)
