#!/usr/bin/env python

###
# Viid
# A network visualizer for software defined networks with the ability to define custom datapaths between nodes
###

from PyQt5 import QtCore, QtGui, QtWidgets

import enums

class Sidebar(QtWidgets.QDockWidget):
    def __init__(self):
        super().__init__()

        self.setAllowedAreas(QtCore.Qt.RightDockWidgetArea)
        self.setFeatures(QtWidgets.QDockWidget.NoDockWidgetFeatures) 

        self.button = QtWidgets.QPushButton()
        self.button.setMaximumWidth(30)

        self.tabs = QtWidgets.QTabWidget()
        self.tabs.setMinimumWidth(400)

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0,0,0,0)
        layout.addWidget(self.button)
        layout.addWidget(self.tabs)

        container = QtWidgets.QWidget()
        container.setLayout(layout)
        self.setWidget(container)

        self.layout = layout

        self.tabs.hide()
        self.toggle_sidebar()

        ### PANEL CONNECTION
        self.connection = Connection()
        ### PANEL DETAIL
        self.detail = Detail()
        ### PANEL PATH DEFINITION
        self.path_definition = PathDefinition()
        ### PANEL SAVE/LOAD
        self.saveload = SaveLoad()

        ### ADD PANELS TO PANEL TAB BAR
        self.tabs.addTab(self.path_definition, "Path definition")
        self.tabs.addTab(self.detail, "Device detail")
        self.tabs.addTab(self.saveload, "Save/load")
        self.tabs.addTab(self.connection, "Connection")

        ### PO OTEVRENI CONNECTION PANEL
        ### taky predelat z widgetu na index?
        self.tabs.setCurrentWidget(self.connection)

    # odemykani a zamykani tabu podle toho, jestli je nebo neni pripojeno
    def connected(self, state):
        self.tabs.setTabEnabled(0, state) # path definition
        self.tabs.setTabEnabled(1, state) # detail
        self.tabs.setTabEnabled(2, state) # save load

    def toggle_sidebar(self, show=False):
        if self.tabs.isHidden() or show:
            self.tabs.show()
            self.button.setText(">>")
            self.setFixedSize(QtWidgets.QWIDGETSIZE_MAX, QtWidgets.QWIDGETSIZE_MAX)
        else:
            self.tabs.hide()
            self.button.setText("<<")
            self.setFixedWidth(37)

    def resize_button(self):
        self.button.setFixedHeight(self.widget().height())


class Connection(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()
        self.form = self.form()

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.form)
        self.setLayout(layout)

    def form(self):
        boldFont = QtGui.QFont('Ubuntu', 12)
        boldFont.setBold(True)

        ip = QtWidgets.QLineEdit()
        username = QtWidgets.QLineEdit()
        password = QtWidgets.QLineEdit()
        self.interface = QtWidgets.QComboBox()

        self.error_message = QtWidgets.QLabel()
        self.error_message.setAlignment(QtCore.Qt.AlignHCenter)
        self.error_message.setStyleSheet("""
            .QLabel {
                font-weight: bold;
                background-color: #ffaaaa;
                padding: 3px 0;
             }
        """)
        self.error_message.hide()
        self.connect = QtWidgets.QPushButton("Connect")
        self.connect.setFont(boldFont)
        self.disconnect = QtWidgets.QPushButton("Disconnect")
        self.disconnect.setEnabled(False)

        form = QtWidgets.QFormLayout()
        form.addRow("Interface:", self.interface)
        form.addRow("Controller IP:", ip)
        form.addRow("Username:", username)
        form.addRow("Password:", password)
        form.addRow(self.error_message)
        form.addRow(self.connect)
        form.addRow(self.disconnect)

        self.form_items = {}
        self.form_items['interface'] = self.interface
        self.form_items['ip'] = ip
        self.form_items['username'] = username
        self.form_items['password'] = password

        form_container = QtWidgets.QWidget()
        form_container.setLayout(form)
        return form_container

    def add_interface(self, name, key):
        self.interface.addItem(name, key)

    def check(self):
        for item in [self.form_items['ip'], self.form_items['username'],
                     self.form_items['password']]:
            if self.is_empty(item):
                self.error_message.setText("All fields must be set")
                self.error_message.show()
                return
        
        self.error_message.hide()
        return True

    def connected(self, state):
        # potreba okomentovat, co co znamena
        for item in self.form_items.values():
            item.setEnabled(not(state))
        self.connect.setEnabled(not(state))
        self.disconnect.setEnabled(state)

    def is_empty(self, field):
        return field.text() == ""


class Detail(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()
        self.table = None

        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)

        self.set_empty()

    def set_datapath(self, name, dpid, ip, of_version):
        self.name = QtWidgets.QLineEdit(name)

        table = QtWidgets.QFormLayout()
        table.addRow("Name:", self.name)
        table.addRow("DPID:", QtWidgets.QLabel(dpid))
        table.addRow("IP address:", QtWidgets.QLabel(ip))
        table.addRow("OpenFlow version:", QtWidgets.QLabel(of_version))

        table_container = QtWidgets.QWidget()
        table_container.setLayout(table)

        self.replace_table(table_container)
       
    def set_node(self, name, mac, ip, parent_dpid, parent_ip, parent_port):
        self.name = QtWidgets.QLineEdit(name)

        table = QtWidgets.QFormLayout()
        table.addRow("Name:", self.name)
        table.addRow("MAC address:", QtWidgets.QLabel(mac))
        table.addRow("IP address:", QtWidgets.QLabel(ip))
        parent_table = QtWidgets.QFormLayout()
        parent_table.addRow("DPID:", QtWidgets.QLabel(parent_dpid))
        parent_table.addRow("IP address:", QtWidgets.QLabel(parent_ip))
        parent_table.addRow("Port:", QtWidgets.QLabel(str(parent_port)))
        
        line = QtWidgets.QFrame()
        line.setFrameShape(QtWidgets.QFrame.HLine)
        table.addRow(line)
        table.addRow(QtWidgets.QLabel("<b>Parent datapath:</b>"))
        table.addRow(parent_table)

        table_container = QtWidgets.QWidget()
        table_container.setLayout(table)

        self.replace_table(table_container)

    def set_empty(self):
        label = QtWidgets.QLabel(
             "Double-click a device to get its details.")
        self.replace_table(label)

    def replace_table(self, new_table):
        if self.table is not None:
            self.table.setParent(None)
        self.layout.addWidget(new_table)
        self.table = new_table


class PathDefinition(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()
        self.table = self.table()

        self.discard = QtWidgets.QPushButton("Discard path")

        self.form = self.form()
        self.form.setEnabled(False)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.table)
        layout.addWidget(self.form)
        layout.addWidget(self.discard)
        self.setLayout(layout)

    def table(self):
        table = QtWidgets.QTableWidget()
        table.setColumnCount(2)
        table.setHorizontalHeaderLabels(['DPID/MAC', 'IP'])
        table.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        table.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        table.horizontalHeader().setStretchLastSection(True)

        return table

    def form(self):
        self.form_items = {}

        name = QtWidgets.QLineEdit()
        priority = QtWidgets.QLineEdit()
        idle_timeout = QtWidgets.QLineEdit()
        hard_timeout = QtWidgets.QLineEdit()
        tcp_src = QtWidgets.QLineEdit()
        tcp_dst = QtWidgets.QLineEdit()
        udp_src = QtWidgets.QLineEdit()
        udp_dst = QtWidgets.QLineEdit()

        self.form_items['name'] = name
        self.form_items['priority'] = priority
        self.form_items['idle_timeout'] = idle_timeout
        self.form_items['hard_timeout'] = hard_timeout
        self.form_items['tcp_src'] = tcp_src
        self.form_items['tcp_dst'] = tcp_dst
        self.form_items['udp_src'] = udp_src
        self.form_items['udp_dst'] = udp_dst

        for i in [tcp_src, tcp_dst, udp_src, udp_dst]:
            i.setEnabled(False)

        tcp = QtWidgets.QRadioButton("TCP")
        tcp.toggled.connect(tcp_src.setEnabled)
        tcp.toggled.connect(tcp_dst.setEnabled)
        udp = QtWidgets.QRadioButton("UDP")
        udp.toggled.connect(udp_src.setEnabled)
        udp.toggled.connect(udp_dst.setEnabled)
        self.all_ip = QtWidgets.QRadioButton("All IP protocols")
        self.all_ip.setChecked(True)

        protocol_button_group = QtWidgets.QButtonGroup()
        protocol_button_group.setExclusive(True)
        protocol_button_group.addButton(tcp, enums.IpProtocol.tcp.value)
        protocol_button_group.addButton(udp, enums.IpProtocol.udp.value)
        protocol_button_group.addButton(self.all_ip,
            enums.IpProtocol.all_ip.value)

        self.form_items['ip_protocol'] = protocol_button_group

        for i in [priority, idle_timeout, hard_timeout,
                  tcp_src, tcp_dst, udp_src, udp_dst]:
            i.setValidator(QtGui.QIntValidator(0, 65535))
            i.setPlaceholderText("0-65535")
        idle_timeout.setPlaceholderText("0-65535 (default 0)")
        hard_timeout.setPlaceholderText("0-65535 (default 0)")
        
        self.error_message = QtWidgets.QLabel()
        self.error_message.setAlignment(QtCore.Qt.AlignHCenter)
        self.error_message.setStyleSheet("""
            .QLabel {
                font-weight: bold;
                background-color: #ffaaaa;
                padding: 3px 0;
             }
        """)
        self.error_message.hide()

        bold_font = QtGui.QFont('Ubuntu', 12)
        bold_font.setBold(True)
        self.confirm = QtWidgets.QPushButton("Define path")
        self.confirm.setFont(bold_font)

        form = QtWidgets.QGridLayout()
        form.addWidget(QtWidgets.QLabel("Path name:"), 0, 0)
        form.addWidget(name, 0, 1)

        form.addWidget(QtWidgets.QLabel("Priority:"), 1, 0)
        form.addWidget(priority, 1, 1)

        form.addWidget(QtWidgets.QLabel("Idle timeout [s]:"), 2, 0)
        form.addWidget(idle_timeout, 2, 1)

        form.addWidget(QtWidgets.QLabel("Hard timeout [s]:"), 3, 0)
        form.addWidget(hard_timeout, 3, 1)

        form.addWidget(tcp, 4, 0)
        form.addWidget(QtWidgets.QLabel("TCP source port:"), 5, 0)
        form.addWidget(tcp_src, 5, 1)
        form.addWidget(QtWidgets.QLabel("TCP destination port:"), 6, 0)
        form.addWidget(tcp_dst, 6, 1)

        form.addWidget(udp, 7, 0)
        form.addWidget(QtWidgets.QLabel("UDP source port:"), 8, 0)
        form.addWidget(udp_src, 8, 1)
        form.addWidget(QtWidgets.QLabel("UDP destination port:"), 9, 0)
        form.addWidget(udp_dst, 9, 1)

        form.addWidget(self.all_ip, 10, 0)

        form.addWidget(self.error_message, 11, 0, 1, 3)
        form.addWidget(self.confirm, 12, 0, 1, 3)

        form_container = QtWidgets.QWidget()
        form_container.setLayout(form)
        return form_container

    def path_complete_state_changed(self, status):
        self.form.setEnabled(status)

    def values(self):
        mandatory = {}
        mandatory['name'] = self.form_items['name'].text()
        mandatory['priority'] = self.form_items['priority'].text()
        mandatory['idle_timeout'] = (self.form_items['idle_timeout'].text()
            if not self.is_empty(self.form_items['idle_timeout']) else 0)
        mandatory['hard_timeout'] = (self.form_items['hard_timeout'].text()
            if not self.is_empty(self.form_items['hard_timeout']) else 0)
        optional = {}
        optional['ip_protocol'] = enums.IpProtocol(
            self.form_items['ip_protocol'].checkedId())
        optional['tcp_src'] = self.form_items['tcp_src'].text()
        optional['tcp_dst'] = self.form_items['tcp_dst'].text()
        optional['udp_src'] = self.form_items['udp_src'].text()
        optional['udp_dst'] = self.form_items['udp_dst'].text()
        return (mandatory, optional)

    def check(self):
        if self.is_empty(self.form_items['name']):
            self.error_message.setText("Path name must be set.")
            self.error_message.show()
            return False
        if self.is_empty(self.form_items['priority']):
            self.error_message.setText("Priority must be set.")
            self.error_message.show()
            return False
        return True
    
    def is_empty(self, field):
        return field.text() == ""
        
    def clear(self):
        for item in self.form_items.values():
            if isinstance(item, QtWidgets.QLineEdit):
                item.clear()
        self.error_message.hide()
        self.all_ip.setChecked(True)


class SaveLoad(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()
        self.form = self.form()

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.form)
        self.setLayout(layout)

    def form(self):
        self.save = QtWidgets.QPushButton("Save")
        self.load = QtWidgets.QPushButton("Load")

        form = QtWidgets.QFormLayout()
        form.addRow(self.save)
        form.addRow(self.load)

        form_container = QtWidgets.QWidget()
        form_container.setLayout(form)
        return form_container
