#!/usr/bin/env python

"""
Module 'factory' provides the means to access available interface modules.
"""

import importlib
import os
import pathlib
import sys

# Dictionary of available interface modules
# module name: interface module full name
listing = {}

def _prepare():
    """Prepare the dictionary of available interface modules.

    This function goes through all .py files present in its dictionary and 
    takes all except those named 'base', 'factory', and '__init__' because it
    is known that these are not files of specific interface modules. A global
    dictionary 'listing' is then filled with pairs of module name and full
    name, which is defined in the module. 
    """
    # Locate the dictionary of this file
    location = os.path.dirname(os.path.realpath(__file__))
    # Take all .py files
    files = pathlib.Path(location+'/').glob('*.py')
    for f in files:
        if f.stem not in {'base', 'factory', '__init__'}:
            # Import the interface module based on the name of the file
            module = importlib.import_module('interfaces.'+f.stem)
            listing[f.stem] = module.FULL_NAME

def create(name):
    """Return an instance of the Network class of a particular module.

    Parameters
    ----------
    name : string
        Name of the particular interface module.

    Returns
    ------
    Network
        A fresh instance of the Network class of the interface module of the
        given name
    """
    # Create a reference to the particular interface module
    # The modules have to be imported in advance (by the _prepare() method)
    interface = sys.modules['interfaces.'+name]
    return interface.Network()
