#!/usr/bin/env python

"""
Module 'layout' provides the transformation between data and its visualization.
"""

import collections
import math

from PyQt5 import QtCore, QtWidgets
import networkx as nx

import data
import gui

Coords = collections.namedtuple('Coords', 'x y')

class Layout:
    """Provides graphical objects and their positioning on the scene."""

    def __init__(self, scene):
        """Initialize instance variables.

        Parameters
        ----------
        scene : gui.scene.Scene
            Reference to the scene to place the graphical elements on.
        """
        self.scene = scene
        self.translation = Translation()
        # Graph structure
        self.graph = nx.Graph()
        self.positions = None

    def clear(self):
        """Clear the layout."""
        self.scene.clear()
        self.translation.clear()
        self.graph.clear()
        self.positions = None

    def create_graphical_element(self, net_element):
        """Create a graphical element for the given network element.

        Parameters
        ----------
        net_element : interfaces.Datapath or interfaces.Node
            Network element to create a graphical counterpart for.

        Returns
        -------
        gui.canvas.Datapath
            In case the input network element was a datapath.
        gui.canvas.Node
            In case the input network element was a node.
        """
        if net_element in self.translation.element_to_graphical:
            return self.translation.element_to_graphical[net_element]

        if isinstance(net_element, data.interfaces.Datapath):
            name = (data.Network.device_names[net_element.dpid]
                if net_element.dpid in data.Network.device_names
                else net_element.dpid)
            level = data.Network.datapath_level(net_element)
            grph_element = self.scene.create_datapath(net_element.ip, name, level)
        elif isinstance(net_element, data.interfaces.Node):
            name = (data.Network.device_names[net_element.mac]
                if net_element.mac in data.Network.device_names
                else net_element.mac)
            grph_element = self.scene.create_node(net_element.ip, name,
                net_element.parent_port)
        
        self.translation.element_to_graphical[net_element] = grph_element
        self.translation.element_from_graphical[grph_element] = net_element

        return grph_element

    def create_graphical_link(self, net_link):
        """Create a graphical link for the given network link.

        Parameters
        ----------
        net_link : interfaces.Link
            Network link to create a graphical counterpart for.
        Returns
        -------
        gui.canvas.Link
        """
        if net_link in self.translation.link_to_graphical:
            return self.translation.link_to_graphical[net_link]

        source = self.translation.element_to_graphical[net_link.src]
        destination = self.translation.element_to_graphical[net_link.dst]
        grph_link = self.scene.create_link(source, destination)

        self.translation.link_to_graphical[net_link] = grph_link

        return grph_link

    def remove_graphical_element(self, net_element):
        """Remove the given network element from translations.

        Parameters
        ----------
        net_element : interfaces.Datapath or interfaces.Node
            Network element to remove.

        Returns
        -------
        gui.canvas.Datapath
            In case the network element is a datapath.
        gui.canvas.Node
            In case the network element is a node.
        None
            In case the network element is not in the dictionary.
        """
        if net_element not in self.translation.element_to_graphical:
            return None

        grph_element = self.translation.element_to_graphical[net_element]
        del self.translation.element_to_graphical[net_element]
        del self.translation.element_from_graphical[grph_element]

        return grph_element

    def remove_graphical_link(self, net_link):
        """Remove the given network link from translations.

        Parameters
        ----------
        net_link : interfaces.Link
            Network link to remove.

        Returns
        -------
        gui.canvas.Link
        None
            In case the network link is not in the dictionary.
        """
        if net_link not in self.translation.link_to_graphical:
            return None

        grph_link = self.translation.link_to_graphical[net_link]
        del self.translation.link_to_graphical[net_link]

        return grph_link

    def graph_create(self, net_elements, net_links):
        """Create a graph structure from the given network data.

        Parameters
        ----------
        net_elements : list of interfaces.Datapath or interfaces.Node
            Datapaths and nodes connected in the network.
        net_links : list of interfaces.Link
            Links connected in the network.
        """
        for net_element in net_elements:
            grph_element = self.create_graphical_element(net_element)
            self.graph.add_node(grph_element)

        grph_links = []
        for net_link in net_links:
            grph_link = self.create_graphical_link(net_link)
            grph_element1 = grph_link.source
            grph_element2 = grph_link.destination
            self.graph.add_edge(grph_element1, grph_element2)
            grph_links.append(grph_link)

        scale = 100 * math.pow(1.15, self.graph.number_of_nodes())
        self.positions = nx.spring_layout(self.graph,
            scale=scale)

        for grph_element in self.positions:
            self.scene.add_element(grph_element, self.positions[grph_element])

        for grph_link in grph_links:
            self.scene.add_link(grph_link)

    def graph_renew(self):
        """Generate new positions for elements present in the graph."""
        self.positions = nx.spring_layout(self.graph,
            scale=(self.graph.number_of_nodes()*70))

        for grph_element in self.positions:
            x, y = self.positions[grph_element]
            grph_element.setPos(x, y)

        for grph_link in self.translation.link_to_graphical.values():
            grph_link.draw()

    def update(self, add, remove, connect, disconnect):
        """Update the visualization.

        Parameters
        ----------
        add : list of interfaces.Datpaath or interfaces.Node
            Network devices that have been added to the network.
        remove : list of interfaces.Datapath or interfaces.Node
            Network devices that have been removed from the network.
        connect : list of interfaces.Link
            Network links that have been connected to the network.
        disconnect : list of interfaces.Link
            Network links that have been disconnected from the network.
        """
        for net_element in remove:
            grph_element = self.remove_graphical_element(net_element)
            self.graph.remove_node(grph_element)
            del self.positions[grph_element]
            self.scene.remove_element(grph_element)

        for net_link in disconnect:
            grph_link = self.remove_graphical_link(net_link)
            self.scene.remove_link(grph_link)

        fixed_elements = self.graph.nodes()

        new_elements = []
        for net_element in add:
            grph_element = self.create_graphical_element(net_element)
            self.graph.add_node(grph_element)
            new_elements.append(grph_element)

        new_links = []
        for net_link in connect:
            grph_link = self.create_graphical_link(net_link)
            grph_element1 = grph_link.source
            grph_element2 = grph_link.destination
            self.graph.add_edge(grph_element1, grph_element2)
            new_links.append(grph_link)

        if self.positions and fixed_elements:
            self.positions = nx.spring_layout(self.graph, pos=self.positions,
                fixed=fixed_elements)
        else:
            self.positions = nx.spring_layout(self.graph,
                scale=(self.graph.number_of_nodes()*70))

        for grph_element in new_elements:
            self.scene.add_element(grph_element, self.positions[grph_element])

        for grph_link in new_links:
            self.scene.add_link(grph_link)

        self.update_stats()

    def update_stats(self):
        """Update link utilization statistics."""
        for net_link, grph_link in self.translation.link_to_graphical.items():
            datapath = net_link.src
            port = datapath.port_of_device(net_link.dst)
            stats = datapath.ports_stats[port]
            # statistiky plati pro cilovy prvek
            # co src vysle(tx) = prijem u dst (rx)
            # proto prohozene v parametrech
            grph_link.update_stats(stats.tx_rate, stats.rx_rate)

    def move_element(self, grph_element, pos):
        """Save the new position for the given graphical element.

        Parameters
        ----------
        grph_element : gui.canvas.Datapath or gui.canvas.Node
            Graphical element that was moved.
        pos : PyQt5.QtCore.QPointF
            New position of the graphical element.
        """
        x = pos.x()
        y = pos.y()
        self.positions[grph_element] = (x, y)

    def tree(self, net_treetop):
        """Transform the visualization into a tree-like structure.

        Parameters
        ----------
        net_treetop : list of interfaces.Datapath
            Datapaths from the top level of the network hierarchy.
        """
        all_grph_elements = self.translation.element_from_graphical.keys()
        tree_grph_elements = []

        top = True
        row = set(net_treetop)
        passed = set()
        next_row = set()
        predecessors = {}
        tree_roots = []
        while len(row) != 0:
            net_datapath = row.pop()  # get a particular Datapath object
            passed.add(net_datapath)

            grph_datapath = self.translation.element_to_graphical[net_datapath]
            tree_datapath = TreeDatapath(grph_datapath)
            tree_grph_elements.append(grph_datapath)
            if top:
                tree_roots.append(tree_datapath)

            for device in net_datapath.ports_devices.values():
                if isinstance(device, data.interfaces.Datapath):
                    if device not in (passed | row | next_row):
                        next_row.add(device)
                        predecessors[device] = tree_datapath
                elif isinstance(device, data.interfaces.Node):
                    grph_node = self.translation.element_to_graphical[device]
                    tree_node = TreeNode(grph_node)
                    tree_datapath.nodes.append(tree_node)
                    tree_grph_elements.append(grph_node)

            if net_datapath in predecessors:
                tree_datapath.predecessor = predecessors[net_datapath]
                predecessors[net_datapath].successors.append(tree_datapath)

            if len(row) == 0:
                top = False
                row = next_row
                next_row = set()

        pos = Coords(0, 0)
        for root in tree_roots:
            pos = self.tree_draw(root, pos)

        remaining = all_grph_elements - tree_grph_elements
        if remaining:
            self.positions = nx.spring_layout(self.graph, pos=self.positions,
                fixed=tree_grph_elements)

            for grph_element in remaining:
                x, y = self.positions[grph_element]
                grph_element.setPos(x, y)
            
        for grph_link in self.translation.link_to_graphical.values():
            grph_link.draw()

    def tree_draw(self, root, start_pos):
        """Draw the tree-like structure using a depth-first search.

        Parameters
        ----------
        root : TreeDatapath or TreeNode
            Root element for the search.
        start_pos: Coords
            Initial coordinates for the element.
        Returns
        -------
        Coords
            End coordinates of the sub-tree.
        """
        if isinstance(root, TreeDatapath):
            width, height = gui.scene.SceneMatrix.element_matrix_size(
                root.grph_element)

            succ_y = start_pos.y + height + 7
            succ_pos = Coords(start_pos.x, succ_y)
            # define end_x in case there are no successors
            end_x = start_pos.x + width
            for succ in root.successors:
                succ_pos = self.tree_draw(succ, succ_pos)
                end_x = max(end_x, succ_pos.x)
            end_x = end_x - 5 if end_x > (start_pos.x + width) else end_x

            first = succ_pos
            for i, node in enumerate(root.nodes, start=1):
                succ_pos = self.tree_draw(node, succ_pos)
                end_x = max(end_x, succ_pos.x)
                if i % 3 == 0:
                    node_width, node_height = (
                        gui.scene.SceneMatrix.element_matrix_size(
                        node.grph_element))
                    new_y = succ_pos.y + node_height + 3
                    succ_pos = Coords(first.x, new_y)

            # vycentrovani nadrazeneho prvku v ramci jeho sekce
            offset_x = int(math.floor((end_x - start_pos.x - width)/2))
            position = Coords(start_pos.x + offset_x, start_pos.y)

            new_scene_pos = gui.scene.SceneMatrix.pos_to_scene(position)
            root.grph_element.setPos(new_scene_pos.x, new_scene_pos.y)
            self.positions[root.grph_element] = (new_scene_pos.x,
                new_scene_pos.y)

            end_pos = Coords(end_x + 5, start_pos.y)
            return end_pos
        elif isinstance(root, TreeNode):
            new_scene_pos = gui.scene.SceneMatrix.pos_to_scene(start_pos)
            root.grph_element.setPos(new_scene_pos.x, new_scene_pos.y)
            self.positions[root.grph_element] = (new_scene_pos.x,
                new_scene_pos.y)

            width, height = gui.scene.SceneMatrix.element_matrix_size(
                root.grph_element)

            end_pos = Coords(start_pos.x + width + 1, start_pos.y)
            return end_pos


class TreeDatapath:
    """Tree representation for a datapath."""

    def __init__(self, grph_element):
        """Initialize instance variables.

        Parameters
        ----------
        grph_element : gui.canvas.Datapath
            Graphical element of a datapath.
        """
        self.grph_element = grph_element
        self.predecessor = None  # TreeDatapath or None
        self.successors = []     # list of TreeDatapath
        self.nodes = []          # list of TreeNode


class TreeNode:
    """Tree representation for a node."""

    def __init__(self, grph_element):
        """Initialize instance variables.

        Parameters
        ----------
        grph_element : gui.canvas.Node
            Graphical element of a node.
        """
        self.grph_element = grph_element


class Translation:
    """Keeps relations between network elements and graphical elements."""

    def __init__(self):
        """Initialize instance variables."""
        self.clear()

    def clear(self):
        """Sets all instance variables to empty dictionaries."""
        self.element_to_graphical = {}
        self.element_from_graphical = {}
        self.link_to_graphical = {}
